<?php

/**
 * @file
 * Command-line runner for migrations.
 */

use Drupal\migrate\Plugin\MigrationInterface;
use Drupal\migrate_run\MigrateExecutable;
use Drupal\migrate_run\DrushLogMigrateMessage;

/**
 * Implements hook_drush_command().
 */
function migrate_run_drush_command() {
  $items['migrate-status'] = [
    'description' => 'List all migrations with current status.',
    'options' => [
      'tag' => 'A comma-separated list of migration tags to list. If only --tag is provided, all migrations will be listed, grouped by tags.',
      'names-only' => 'Only return names, not all the details (faster)',
    ],
    'arguments' => [
      'migration' => 'Restrict to a comma-separated list of migrations. Optional',
    ],
    'examples' => [
      'migrate-status' => 'Retrieve status for all migrations',
      'migrate-status --tag' => 'Retrieve status for all migrations, grouped by tag',
      'migrate-status --tag=user,main_content' => 'Retrieve status for all migrations tagged with "user" or "main_content"',
      'migrate-status classification,article' => 'Retrieve status for specific migrations',
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['ms'],
  ];

  $items['migrate-import'] = [
    'description' => 'Perform one or more migration processes.',
    'options' => [
      'all' => 'Process all migrations.',
      'tag' => 'A comma-separated list of migration tags to import',
      'limit' => 'Limit on the number of items to process in each migration',
      'feedback' => 'Frequency of progress messages, in items processed',
      'idlist' => 'Comma-separated list of IDs to import',
      'update' => 'In addition to processing unprocessed items from the source, update previously-imported items with the current data',
      'force' => 'Force an operation to run, even if all dependencies are not satisfied',
      'execute-dependencies' => 'Execute all dependent migrations first.',
      'timestamp' => 'Show progress ending timestamp in progress messages.',
      'total' => 'Show total processed item number in progress messages.',
    ],
    'arguments' => [
      'migration' => 'ID of migration(s) to import. Delimit multiple using commas.',
    ],
    'examples' => [
      'migrate-import --all' => 'Perform all migrations',
      'migrate-import --tag=user,main_content' => 'Import all migrations tagged with user and main_content tags',
      'migrate-import classification,article' => 'Import new terms and nodes',
      'migrate-import beer_user --limit=2' => 'Import no more than 2 users',
      'migrate-import beer_user --idlist=5' => 'Import the user record with source ID 5',
      'migrate-import beer_user --limit=50 --feedback=20' => 'Import 50 users and show process message every 20th record'
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['mi'],
  ];

  $items['migrate-rollback'] = [
    'description' => 'Rollback one or more migrations.',
    'options' => [
      'all' => 'Process all migrations.',
      'tag' => 'A comma-separated list of migration tags to rollback',
      'feedback' => 'Frequency of progress messages, in items processed',
    ],
    'arguments' => [
      'migration' => 'Name of migration(s) to rollback. Delimit multiple using commas.',
    ],
    'examples' => [
      'migrate-rollback --all' => 'Perform all migrations',
      'migrate-rollback --tag=user,main_content' => 'Rollback all migrations tagged with user and main_content tags',
      'migrate-rollback classification,article' => 'Rollback imported terms and nodes',
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['mr'],
  ];

  $items['migrate-stop'] = [
    'description' => 'Stop an active migration operation.',
    'arguments' => [
      'migration' => 'ID of migration to stop',
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['mst'],
  ];

  $items['migrate-reset-status'] = [
    'description' => "Reset a active migration's status to idle.",
    'arguments' => [
      'migration' => 'ID of migration to reset',
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['mrs'],
  ];

  $items['migrate-messages'] = [
    'description' => 'View any messages associated with a migration.',
    'arguments' => [
      'migration' => 'ID of the migration',
    ],
    'options' => [
      'csv' => 'Export messages as a CSV'
    ],
    'examples' => [
      'migrate-messages article' => 'Show all messages for the article migration',
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['mmsg'],
  ];

  $items['migrate-fields-source'] = [
    'description' => 'List the fields available for mapping in a source.',
    'arguments' => [
      'migration' => 'ID of the migration',
    ],
    'examples' => [
      'migrate-fields-source article' => 'List fields for the source in the article migration',
    ],
    'drupal dependencies' => ['migrate_run'],
    'aliases' => ['mfs'],
  ];

  return $items;
}

/**
 * Retrieves a list of active migrations.
 *
 * @param string $migration_ids
 *   (optional) A comma-separated list of migration IDs. If omitted, will return
 *   all migrations.
 *
 * @return MigrationInterface[][]
 *   An array keyed by migration tag, each value containing an array of
 *   migrations or an empty array if no migrations match the input criteria.
 */
function drush_migrate_run_migration_list($migration_ids = '') {
  $migration_ids = array_filter(array_map('trim', explode(',', $migration_ids)));

  /** @var \Drupal\migrate\Plugin\MigrationInterface[] $migrations */
  $migrations = \Drupal::service('plugin.manager.migration')
    ->createInstances($migration_ids);

  $tags = drush_get_option('tag');

  // If --tag was not passed, don't group on tags, use a single empty tag.
  if ($tags === NULL) {
    return [NULL => $migrations];
  }

  if ($tags !== TRUE) {
    $tags = array_filter(array_map('trim', explode(',', $tags)));
  }

  $list = [];
  foreach ($migrations as $migration_id => $migration) {
    $migration_tags = (array) $migration->getMigrationTags();
    if ($tags !== TRUE) {
      $common_tags = array_intersect($tags, $migration_tags);
      if ($tags && !$common_tags) {
        // Skip if this a list of tags was passed and this migration is not tagged
        // with any of the requested tag.
        continue;
      }
      $grouping_tags = array_intersect($tags, $migration_tags);
    }
    else {
      $grouping_tags = $migration_tags;
    }

    foreach ($grouping_tags as $tag) {
      $list[$tag][$migration_id] = $migration;
    }
  }

  // Sort the list.
  ksort($list);

  return $list;
}

/**
 * @param string $migration_ids
 *   Comma-separated list of migration IDs.
 */
function drush_migrate_run_migrate_status($migration_ids = '') {
  $names_only = drush_get_option('names-only');

  $list = drush_migrate_run_migration_list($migration_ids);

  $table = $cache = [];
  // Take it one tag at a time, listing the migrations within each tag.
  foreach ($list as $tag => $migrations) {
    ksort($migrations);
    $tag_label = $tag ? dt('Tag: @name', ['@name' => $tag]) : dt('Migration');
    if ($names_only) {
      $table[] = [$tag_label];
    }
    else {
      $table[] = [
        $tag_label,
        dt('Status'),
        dt('Total'),
        dt('Imported'),
        dt('Unprocessed'),
      ];
    }
    foreach ($migrations as $migration_id => $migration) {
      if (!isset($cache[$migration_id])) {
        if ($names_only) {
          $row = [$migration_id];
        }
        else {
          try {
            $map = $migration->getIdMap();
            $imported = $map->importedCount();
            $source_plugin = $migration->getSourcePlugin();
          }
          catch (Exception $e) {
            drush_log(dt('Failure retrieving information on @migration: @message',
              ['@migration' => $migration_id, '@message' => $e->getMessage()]));
            continue;
          }
          try {
            $source_rows = $source_plugin->count();
            // -1 indicates uncountable sources.
            if ($source_rows == -1) {
              $source_rows = dt('N/A');
              $unprocessed = dt('N/A');
            }
            else {
              $unprocessed = $source_rows - $map->processedCount();
              if ($source_rows > 0 && $imported > 0) {
                $imported .= ' (' . round(($imported / $source_rows) * 100, 1) . '%)';
              }
            }
          }
          catch (Exception $e) {
            drush_print($e->getMessage());
            drush_log(dt('Could not retrieve source count from @migration: @message',
              ['@migration' => $migration_id, '@message' => $e->getMessage()]));
            $source_rows = dt('N/A');
            $unprocessed = dt('N/A');
          }
          $status = $migration->getStatusLabel();
          $row = [$migration_id, $status, $source_rows, $imported, $unprocessed];
        }
        $cache[$migration_id] = $row;
      }
      $table[] = $cache[$migration_id];
    }
    $table[] = array_fill(0, 5, '');
  }
  drush_print_table($table);
}

/**
 * @param string $migration_names
 *   Comma-separated list of migration IDs.
 */
function drush_migrate_run_migrate_import($migration_names = '') {
  $tags = drush_get_option('tag');
  $all = drush_get_option('all');

  $options = [];
  if (!$all && !$migration_names && !$tags) {
    drush_set_error('MIGRATE_ERROR', dt('You must specify --all, --tag or one or more migration names separated by commas'));
    return;
  }

  foreach (['limit', 'feedback', 'idlist', 'update', 'force', 'timestamp', 'total'] as $option) {
    if (drush_get_option($option)) {
      $options[$option] = drush_get_option($option);
    }
  }

  $list = drush_migrate_run_migration_list($migration_names);
  if (empty($list)) {
    drush_log(dt('No migrations found.'), 'error');
  }

  // Take it one group at a time, importing the migrations within each group.
  foreach ($list as $tag => $migrations) {
    array_walk($migrations, '_drush_migrate_run_execute_migration', $options);
  }
}

/**
 * Executes a single migration. If the --execute-dependencies option was given,
 * the migration's dependencies will also be executed first.
 *
 * @param \Drupal\migrate\Plugin\MigrationInterface $migration
 *  The migration to execute.
 * @param string $migration_id
 *  The migration ID (not used, just an artifact of array_walk()).
 * @param array $options
 *  Additional options for the migration.
 */
function _drush_migrate_run_execute_migration(MigrationInterface $migration, $migration_id,  array $options = []) {
  $log = new DrushLogMigrateMessage();

  if (drush_get_option('execute-dependencies')) {
    $dependencies = $migration->getMigrationDependencies();
    $required_ids = isset($dependencies['required']) ? $dependencies['required'] : NULL;
    if ($required_ids) {
      $required_migrations = \Drupal::service('plugin.manager.migration')
        ->createInstances($required_ids);
      $dependency_options = array_merge($options, ['is_dependency' => TRUE]);
      array_walk($required_migrations, __FUNCTION__, $dependency_options);
    }
  }
  if (!empty($options['force'])) {
    $migration->set('requirements', []);
  }
  if (!empty($options['update'])) {
    $migration->getIdMap()->prepareUpdate();
  }
  $executable = new MigrateExecutable($migration, $log, $options);
  // drush_op() provides --simulate support
  drush_op(array($executable, 'import'));
}

/**
 * @param string $migration_ids
 */
function drush_migrate_run_migrate_rollback($migration_ids = '') {
  $tags = drush_get_option('tag');
  $all = drush_get_option('all');
  $options = [];
  if (!$all && !$migration_ids && !$tags) {
    drush_set_error('MIGRATE_ERROR', dt('You must specify --all, --tag, or one or more migration names separated by commas'));
    return;
  }

  if (drush_get_option('feedback')) {
    $options['feedback'] = drush_get_option('feedback');
  }

  $log = new DrushLogMigrateMessage();

  $list = drush_migrate_run_migration_list($migration_ids);
  if (empty($list)) {
    drush_log(dt('No migrations found.'), 'error');
  }

  // Take it one group at a time, rolling back the migrations within each group.
  foreach ($list as $group_id => $migrations) {
    // Roll back in reverse order.
    $migrations = array_reverse($migrations);
    foreach ($migrations as $migration_id => $migration) {
      $executable = new MigrateExecutable($migration, $log, $options);
      // drush_op() provides --simulate support.
      drush_op(array($executable, 'rollback'));
    }
  }
}

/**
 * @param string $migration_id
 */
function drush_migrate_run_migrate_stop($migration_id = '') {
  /** @var MigrationInterface $migration */
  $migration = \Drupal::service('plugin.manager.migration')->createInstance($migration_id);
  if ($migration) {
    $status = $migration->getStatus();
    switch ($status) {
      case MigrationInterface::STATUS_IDLE:
        drush_log(dt('Migration @id is idle', ['@id' => $migration_id]), 'warning');
        break;
      case MigrationInterface::STATUS_DISABLED:
        drush_log(dt('Migration @id is disabled', ['@id' => $migration_id]), 'warning');
        break;
      case MigrationInterface::STATUS_STOPPING:
        drush_log(dt('Migration @id is already stopping', ['@id' => $migration_id]), 'warning');
        break;
      default:
        $migration->interruptMigration(MigrationInterface::RESULT_STOPPED);
        drush_log(dt('Migration @id requested to stop', ['@id' => $migration_id]), 'success');
        break;
    }
  }
  else {
    drush_log(dt('Migration @id does not exist', ['@id' => $migration_id]), 'error');
  }
}

/**
 * @param string $migration_id
 */
function drush_migrate_run_migrate_reset_status($migration_id = '') {
  /** @var MigrationInterface $migration */
  $migration = \Drupal::service('plugin.manager.migration')->createInstance($migration_id);
  if ($migration) {
    $status = $migration->getStatus();
    if ($status == MigrationInterface::STATUS_IDLE) {
      drush_log(dt('Migration @id is already Idle', ['@id' => $migration_id]), 'warning');
    }
    else {
      $migration->setStatus(MigrationInterface::STATUS_IDLE);
      drush_log(dt('Migration @id reset to Idle', ['@id' => $migration_id]), 'status');
    }
  }
  else {
    drush_log(dt('Migration @id does not exist', ['@id' => $migration_id]), 'error');
  }
}

/**
 * @param string $migration_id
 */
function drush_migrate_run_migrate_messages($migration_id) {
  /** @var MigrationInterface $migration */
  $migration = \Drupal::service('plugin.manager.migration')->createInstance($migration_id);
  if ($migration) {
    $map = $migration->getIdMap();
    $first = TRUE;
    $table = [];
    foreach ($map->getMessageIterator() as $row) {
      unset($row->msgid);
      if ($first) {
        // @todo: Ideally, replace sourceid* with source key names. Or, should
        // getMessageIterator() do that?
        foreach ($row as $column => $value) {
          $table[0][] = $column;
        }
        $first = FALSE;
      }
      $table[] = (array)$row;
    }
    if (empty($table)) {
      drush_log(dt('No messages for this migration'), 'status');
    }
    else {
      if (drush_get_option('csv')) {
        foreach ($table as $row) {
          fputcsv(STDOUT, $row);
        }
      }
      else {
        $widths = [];
        foreach ($table[0] as $header) {
          $widths[] = strlen($header) + 1;
        }
        drush_print_table($table, TRUE, $widths);
      }
    }
  }
  else {
    drush_log(dt('Migration @id does not exist', ['@id' => $migration_id]), 'error');
  }
}

/**
 * @param string $migration_id
 */
function drush_migrate_run_migrate_fields_source($migration_id) {
  /** @var MigrationInterface $migration */
  $migration = \Drupal::service('plugin.manager.migration')->createInstance($migration_id);
  if ($migration) {
    $source = $migration->getSourcePlugin();
    $table = [];
    foreach ($source->fields() as $machine_name => $description) {
      $table[] = [strip_tags($description), $machine_name];
    }
    drush_print_table($table);
  }
  else {
    drush_log(dt('Migration @id does not exist', ['@id' => $migration_id]), 'error');
  }
}
